# Copyright (C) 2016-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import locale
from typing import TYPE_CHECKING, Iterator, List, Optional

import click

from . import cli

if TYPE_CHECKING:
    import datetime

    # importing swh.storage.interface triggers the load of 300+ modules, so...
    import swh.model.model
    from swh.storage.interface import StorageInterface


locale.setlocale(locale.LC_ALL, "")
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
DATETIME = click.DateTime()


@cli.group("task")
@click.pass_context
def task(ctx):
    """Manipulate tasks.

    Expected configuration:

    \b
    * :ref:`cli-config-scheduler`
    """
    if not ctx.obj["scheduler"]:
        ctx.fail("Scheduler class (local/remote) must be instantiated")


@task.command("schedule")
@click.option(
    "--columns",
    "-c",
    multiple=True,
    default=["type", "args", "kwargs", "next_run"],
    type=click.Choice(["type", "args", "kwargs", "policy", "next_run"]),
    help="columns present in the CSV file",
)
@click.option("--delimiter", "-d", default=",")
@click.argument("file", type=click.File(encoding="utf-8"))
@click.pass_context
def schedule_tasks(ctx, columns, delimiter, file):
    """Schedule tasks from a CSV input file.

    The following columns are expected, and can be set through the -c option:

    \b
    - type: the type of the task to be scheduled (mandatory)
    - args: the arguments passed to the task (JSON list, defaults to an empty
      list)
    - kwargs: the keyword arguments passed to the task (JSON object, defaults
      to an empty dict)
    - next_run: the date at which the task should run (datetime, defaults to
      now)

    The CSV can be read either from a named file, or from stdin (use - as
    filename).

    Use sample::

        \b
        cat scheduling-task.txt | \\
            python3 -m swh.scheduler.cli \\
                --database 'service=swh-scheduler-dev' \\
                task schedule \\
                    --columns type --columns kwargs --columns policy \\
                    --delimiter ';' -
    """
    import csv
    import json

    from swh.scheduler.utils import utcnow

    from .utils import pretty_print_task

    tasks = []
    now = utcnow()
    scheduler = ctx.obj["scheduler"]

    reader = csv.reader(file, delimiter=delimiter)
    for line in reader:
        task = dict(zip(columns, line))
        args = json.loads(task.pop("args", "[]"))
        kwargs = json.loads(task.pop("kwargs", "{}"))
        task["arguments"] = {
            "args": args,
            "kwargs": kwargs,
        }
        task["next_run"] = task.get("next_run", now)
        tasks.append(task)

    created = scheduler.create_tasks(tasks)

    output = [
        "Created %d tasks\n" % len(created),
    ]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo_via_pager("\n".join(output))


@task.command("add")
@click.argument("task_type_name", nargs=1, required=True)
@click.argument("options", nargs=-1)
@click.option(
    "--policy", "-p", default="recurring", type=click.Choice(["recurring", "oneshot"])
)
@click.option(
    "--priority", "-P", default=None, type=click.Choice(["low", "normal", "high"])
)
@click.option("--next-run", "-n", default=None)
@click.pass_context
def schedule_task(ctx, task_type_name, options, policy, priority, next_run):
    """Schedule one task from arguments.

    The first argument is the name of the task type. Flag options (policy, priority) are
    task configuration. Further options are positional and keyword argument(s) of the
    task, in YAML format. Keyword args are of the form key=value.

    Usage sample::

        \b
        swh-scheduler --database 'service=swh-scheduler' \\
            task add list-pypi

        \b
        swh-scheduler --database 'service=swh-scheduler' \\
            task add list-debian-distribution --policy=oneshot distribution=stretch

    Note: if the priority is not given, the task won't have the priority set,
    which is considered as the lowest priority level.

    """
    from .utils import parse_options, task_add

    scheduler = ctx.obj["scheduler"]

    task_type = scheduler.get_task_type(task_type_name)
    if not task_type:
        ctx.fail(f"Unknown task name {task_type_name}.")

    (args, kw) = parse_options(options)
    task_add(
        scheduler,
        task_type_name=task_type_name,
        policy=policy,
        priority=priority,
        next_run=next_run,
        args=args,
        kw=kw,
    )


def iter_origins(  # use string annotations to prevent some pkg loading
    storage: StorageInterface,
    page_token: Optional[str] = None,
) -> Iterator[swh.model.model.Origin]:
    """Iterate over origins in the storage. Optionally starting from page_token.

    This logs regularly an info message during pagination with the page_token. This, in
    order to feed it back to the cli if the process interrupted.

    Yields
        origin model objects from the storage

    """
    while True:
        page_result = storage.origin_list(page_token=page_token)
        page_token = page_result.next_page_token
        yield from page_result.results
        if not page_token:
            break
        click.echo(f"page_token: {page_token}\n")


@task.command("schedule_origins")
@click.argument("type", nargs=1, required=True)
@click.argument("options", nargs=-1)
@click.option(
    "--batch-size",
    "-b",
    "origin_batch_size",
    default=10,
    show_default=True,
    type=int,
    help="Number of origins per task",
)
@click.option(
    "--page-token",
    default=0,
    show_default=True,
    type=str,
    help="Only schedule tasks for origins whose ID is greater",
)
@click.option(
    "--limit",
    default=None,
    type=int,
    help="Limit the tasks scheduling up to this number of tasks",
)
@click.option("--storage-url", "-g", help="URL of the (graph) storage API")
@click.option(
    "--dry-run/--no-dry-run",
    is_flag=True,
    default=False,
    help="List only what would be scheduled.",
)
@click.pass_context
def schedule_origin_metadata_index(
    ctx, type, options, storage_url, origin_batch_size, page_token, limit, dry_run
):
    """Schedules tasks for origins that are already known.

    The first argument is the name of the task type, further ones are
    keyword argument(s) of the task in the form key=value, where value is
    in YAML format.

    Usage sample::

        \b
        swh-scheduler --database 'service=swh-scheduler' \\
            task schedule_origins index-origin-metadata
    """
    from itertools import islice

    from swh.storage import get_storage

    from .utils import parse_options, schedule_origin_batches

    scheduler = ctx.obj["scheduler"]
    storage = get_storage("remote", url=storage_url)
    if dry_run:
        scheduler = None

    (args, kw) = parse_options(options)
    if args:
        raise click.ClickException("Only keywords arguments are allowed.")

    origins = iter_origins(storage, page_token=page_token)
    if limit:
        origins = islice(origins, limit)

    origin_urls = (origin.url for origin in origins)
    schedule_origin_batches(scheduler, type, origin_urls, origin_batch_size, kw)


@task.command("list-pending")
@click.argument("task-types", required=True, nargs=-1)
@click.option(
    "--limit",
    "-l",
    "num_tasks",
    required=False,
    type=click.INT,
    help="The maximum number of tasks to fetch",
)
@click.option(
    "--before",
    "-b",
    required=False,
    type=DATETIME,
    help="List all jobs supposed to run before the given date",
)
@click.pass_context
def list_pending_tasks(ctx, task_types, num_tasks, before):
    """List tasks with no priority that are going to be run.

    You can override the number of tasks to fetch with the --limit flag.

    """
    from .utils import pretty_print_task

    scheduler = ctx.obj["scheduler"]

    output = []
    for task_type in task_types:
        pending = scheduler.peek_ready_tasks(
            task_type,
            timestamp=before,
            num_tasks=num_tasks,
        )
        output.append("Found %d %s tasks\n" % (len(pending), task_type))

        for task in pending:
            output.append(pretty_print_task(task))

    click.echo("\n".join(output))


@task.command("list")
@click.option(
    "--task-id",
    "-i",
    default=None,
    multiple=True,
    metavar="ID",
    help="List only tasks whose id is ID.",
)
@click.option(
    "--task-type",
    "-t",
    default=None,
    multiple=True,
    metavar="TYPE",
    help="List only tasks of type TYPE",
)
@click.option(
    "--limit",
    "-l",
    required=False,
    type=click.INT,
    help="The maximum number of tasks to fetch.",
)
@click.option(
    "--status",
    "-s",
    multiple=True,
    metavar="STATUS",
    type=click.Choice(
        ("next_run_not_scheduled", "next_run_scheduled", "completed", "disabled")
    ),
    default=None,
    help="List tasks whose status is STATUS.",
)
@click.option(
    "--policy",
    "-p",
    default=None,
    type=click.Choice(["recurring", "oneshot"]),
    help="List tasks whose policy is POLICY.",
)
@click.option(
    "--priority",
    "-P",
    default=None,
    multiple=True,
    type=click.Choice(["all", "low", "normal", "high"]),
    help="List tasks whose priority is PRIORITY.",
)
@click.option(
    "--before",
    "-b",
    required=False,
    type=DATETIME,
    metavar="DATETIME",
    help="Limit to tasks supposed to run before the given date.",
)
@click.option(
    "--after",
    "-a",
    required=False,
    type=DATETIME,
    metavar="DATETIME",
    help="Limit to tasks supposed to run after the given date.",
)
@click.option(
    "--list-runs",
    "-r",
    is_flag=True,
    default=False,
    help="Also list past executions of each task.",
)
@click.pass_context
def list_tasks(
    ctx, task_id, task_type, limit, status, policy, priority, before, after, list_runs
):
    """List tasks."""
    from operator import itemgetter

    from .utils import pretty_print_run, pretty_print_task

    scheduler = ctx.obj["scheduler"]

    if not task_type:
        task_type = [x["type"] for x in scheduler.get_task_types()]

    # if task_id is not given, default value for status is
    #  'next_run_not_scheduled'
    # if task_id is given, default status is 'all'
    if task_id is None and status is None:
        status = ["next_run_not_scheduled"]
    if status and "all" in status:
        status = None

    if priority and "all" in priority:
        priority = None

    output = []
    tasks = scheduler.search_tasks(
        task_id=task_id,
        task_type=task_type,
        status=status,
        priority=priority,
        policy=policy,
        before=before,
        after=after,
        limit=limit,
    )
    if list_runs:
        runs = {t["id"]: [] for t in tasks}
        for r in scheduler.get_task_runs([task["id"] for task in tasks]):
            runs[r["task"]].append(r)
    else:
        runs = {}

    output.append("Found %d tasks\n" % (len(tasks)))
    for task in sorted(tasks, key=itemgetter("id")):
        output.append(pretty_print_task(task, full=True))
        if runs.get(task["id"]):
            output.append(click.style("  Executions:", bold=True))
            for run in sorted(runs[task["id"]], key=itemgetter("id")):
                output.append(pretty_print_run(run, indent=4))

    click.echo("\n".join(output))


@task.command("respawn")
@click.argument("task-ids", required=True, nargs=-1)
@click.option(
    "--next-run",
    "-n",
    required=False,
    type=DATETIME,
    metavar="DATETIME",
    default=None,
    help="Re spawn the selected tasks at this date",
)
@click.pass_context
def respawn_tasks(ctx, task_ids: List[str], next_run: datetime.datetime):
    """Respawn tasks.

    Respawn tasks given by their ids (see the 'task list' command to
    find task ids) at the given date (immediately by default).

    For example::

        \b
        swh-scheduler task respawn 1 3 12
    """
    from swh.scheduler.utils import utcnow

    scheduler = ctx.obj["scheduler"]
    if next_run is None:
        next_run = utcnow()
    output = []

    task_ids_int = [int(id_) for id_ in task_ids]

    scheduler.set_status_tasks(
        task_ids_int, status="next_run_not_scheduled", next_run=next_run
    )
    output.append("Respawn tasks %s\n" % (task_ids_int,))

    click.echo("\n".join(output))



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
import uuid

import pytest

from swh.scheduler import model, utils

from .common import LISTERS


@pytest.fixture
def mock_datetime(mocker):
    now = datetime.now(tz=timezone.utc)
    mock_datetime = mocker.patch("swh.scheduler.utils.datetime")
    mock_datetime.now.return_value = now
    return mock_datetime


def test_create_oneshot_task_dict_simple(mock_datetime):
    actual_task = utils.create_oneshot_task_dict("some-task-type")

    expected_task = {
        "policy": "oneshot",
        "type": "some-task-type",
        "next_run": mock_datetime.now.return_value,
        "arguments": {
            "args": [],
            "kwargs": {},
        },
    }

    assert actual_task == expected_task
    mock_datetime.now.assert_called_once_with(tz=timezone.utc)


def test_create_oneshot_task_dict_with_next_run():
    next_run = datetime.now(tz=timezone.utc) + timedelta(hours=1)
    actual_task = utils.create_oneshot_task_dict("some-task-type", next_run=next_run)

    expected_task = {
        "policy": "oneshot",
        "type": "some-task-type",
        "next_run": next_run,
        "arguments": {
            "args": [],
            "kwargs": {},
        },
    }

    assert actual_task == expected_task


def test_create_oneshot_task_dict_other_call(mock_datetime):
    actual_task = utils.create_oneshot_task_dict(
        "some-task-type", "arg0", "arg1", priority="high", other_stuff="normal"
    )

    expected_task = {
        "policy": "oneshot",
        "type": "some-task-type",
        "next_run": mock_datetime.now.return_value,
        "arguments": {
            "args": ("arg0", "arg1"),
            "kwargs": {"other_stuff": "normal"},
        },
        "priority": "high",
    }

    assert actual_task == expected_task
    mock_datetime.now.assert_called_once_with(tz=timezone.utc)


def test_create_task_dict(mock_datetime):
    actual_task = utils.create_task_dict(
        "task-type",
        "recurring",
        "arg0",
        "arg1",
        priority="low",
        other_stuff="normal",
        retries_left=3,
    )

    expected_task = {
        "policy": "recurring",
        "type": "task-type",
        "next_run": mock_datetime.now.return_value,
        "arguments": {
            "args": ("arg0", "arg1"),
            "kwargs": {"other_stuff": "normal"},
        },
        "priority": "low",
        "retries_left": 3,
    }

    assert actual_task == expected_task
    mock_datetime.now.assert_called_once_with(tz=timezone.utc)


def test_create_origin_task_dict():
    lister = model.Lister(**LISTERS[1], id=uuid.uuid4())
    origin = model.ListedOrigin(
        lister_id=lister.id,
        url="http://example.com/",
        visit_type="git",
    )

    task = utils.create_origin_task_dict(origin, lister)
    assert task == {
        "type": "load-git",
        "arguments": {
            "args": [],
            "kwargs": {
                "url": "http://example.com/",
                "lister_name": LISTERS[1]["name"],
                "lister_instance_name": LISTERS[1]["instance_name"],
            },
        },
    }

    loader_args = {"foo": "bar", "baz": {"foo": "bar"}}

    origin_w_args = model.ListedOrigin(
        lister_id=lister.id,
        url="http://example.com/svn/",
        visit_type="svn",
        extra_loader_arguments=loader_args,
    )

    task_w_args = utils.create_origin_task_dict(origin_w_args, lister)
    assert task_w_args == {
        "type": "load-svn",
        "arguments": {
            "args": [],
            "kwargs": {
                "url": "http://example.com/svn/",
                "lister_name": LISTERS[1]["name"],
                "lister_instance_name": LISTERS[1]["instance_name"],
                **loader_args,
            },
        },
    }


def test_create_origin_task_dicts(swh_scheduler):
    listers = []
    for lister_args in LISTERS:
        listers.append(swh_scheduler.get_or_create_lister(**lister_args))

    origin1 = model.ListedOrigin(
        lister_id=listers[0].id,
        url="http://example.com/1",
        visit_type="git",
    )
    origin2 = model.ListedOrigin(
        lister_id=listers[0].id,
        url="http://example.com/2",
        visit_type="git",
    )
    origin3 = model.ListedOrigin(
        lister_id=listers[1].id,
        url="http://example.com/3",
        visit_type="git",
    )

    origins = [origin1, origin2, origin3]

    tasks = utils.create_origin_task_dicts(origins, swh_scheduler)
    assert tasks == [
        {
            "type": "load-git",
            "arguments": {
                "args": [],
                "kwargs": {
                    "url": "http://example.com/1",
                    "lister_name": LISTERS[0]["name"],
                    "lister_instance_name": LISTERS[0]["instance_name"],
                },
            },
        },
        {
            "type": "load-git",
            "arguments": {
                "args": [],
                "kwargs": {
                    "url": "http://example.com/2",
                    "lister_name": LISTERS[0]["name"],
                    "lister_instance_name": LISTERS[0]["instance_name"],
                },
            },
        },
        {
            "type": "load-git",
            "arguments": {
                "args": [],
                "kwargs": {
                    "url": "http://example.com/3",
                    "lister_name": LISTERS[1]["name"],
                    "lister_instance_name": LISTERS[1]["instance_name"],
                },
            },
        },
    ]



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta

import pytest

from swh.scheduler.cli.utils import lister_task_type, parse_time_interval


@pytest.mark.parametrize(
    "lister_name, lister_type, expected_task_type",
    [
        ["cgit", None, "list-cgit"],
        ["gitlab", "full", "list-gitlab-full"],
        ["gitea", "incremental", "list-gitea-incremental"],
    ],
)
def test_lister_task_type(lister_name, lister_type, expected_task_type):
    assert lister_task_type(lister_name, lister_type) == expected_task_type


@pytest.mark.parametrize(
    "time_str,expected_timedelta",
    [
        ("1 day", timedelta(days=1)),
        ("1 days", timedelta(days=1)),
        ("2.5 hours", timedelta(hours=2.5)),
        ("99h", timedelta(hours=99)),
        ("10day 2hr 1min 22sec", timedelta(days=10, hours=2, minutes=1, seconds=22)),
        ("8d3h33m04s", timedelta(days=8, hours=3, minutes=33, seconds=4)),
    ],
)
def test_parse_time_interval(time_str, expected_timedelta):
    assert parse_time_interval(time_str) == expected_timedelta


def test_parse_time_interval_raise():
    # Misordered time interval (seconds before days here) is not supported
    with pytest.raises(ValueError, match="not be parsed"):
        parse_time_interval("10s 11 days")

    # Empty string is not parseable
    with pytest.raises(ValueError, match="not be parsed"):
        parse_time_interval("")



# Copyright (C) 2016-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import logging

import click

from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup
from swh.core.cli import swh as swh_cli_group

# If you're looking for subcommand imports, they are further down this file to
# avoid a circular import!


@swh_cli_group.group(
    name="scheduler", context_settings=CONTEXT_SETTINGS, cls=AliasedGroup
)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.option(
    "--database",
    "-d",
    default=None,
    help="Scheduling database DSN (imply cls is 'postgresql')",
)
@click.option(
    "--url", "-u", default=None, help="Scheduler's url access (imply cls is 'remote')"
)
@click.option(
    "--no-stdout", is_flag=True, default=False, help="Do NOT output logs on the console"
)
@click.pass_context
def cli(ctx, config_file, database, url, no_stdout):
    """Software Heritage Scheduler tools.

    Use a local scheduler instance by default (plugged to the
    main scheduler db).

    Expected configuration:

    * :ref:`cli-config-scheduler`
    """
    try:
        from psycopg2 import OperationalError
    except ImportError:

        class OperationalError(Exception):
            pass

    from swh.core import config
    from swh.scheduler import DEFAULT_CONFIG, get_scheduler

    ctx.ensure_object(dict)

    logger = logging.getLogger(__name__)
    scheduler = None
    conf = config.read(config_file, DEFAULT_CONFIG)
    if "scheduler" not in conf:
        ctx.fail("missing 'scheduler' configuration")

    if database:
        conf["scheduler"]["cls"] = "postgresql"
        conf["scheduler"]["db"] = database
    elif url:
        conf["scheduler"]["cls"] = "remote"
        conf["scheduler"]["url"] = url
    sched_conf = conf["scheduler"]
    try:
        logger.debug("Instantiating scheduler with %s", sched_conf)
        scheduler = get_scheduler(**sched_conf)
    except (ValueError, OperationalError) as e:
        # Propagate scheduler instantiation exception context to subcommands, and let
        # them report properly the issue
        ctx.obj["scheduler_exc"] = e

    ctx.obj["scheduler"] = scheduler
    ctx.obj["config"] = conf


from . import (  # noqa
    add_forge_now,
    admin,
    celery_monitor,
    journal,
    origin,
    simulator,
    task,
    task_type,
)


def main():
    import click.core

    click.core.DEPRECATED_HELP_NOTICE = """

DEPRECATED! Please use the command 'swh scheduler'."""
    cli.deprecated = True
    return cli(auto_envvar_prefix="SWH_SCHEDULER")


if __name__ == "__main__":
    main()



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

__all__ = [
    "SchedulerException",
    "StaleData",
    "UnknownPolicy",
]


class SchedulerException(Exception):
    pass


class StaleData(SchedulerException):
    pass


class UnknownPolicy(SchedulerException):
    pass



# Copyright (C) 2023 the Software Heritage developers
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import shared_task


@shared_task(name=__name__ + ".BarLoaderTask")
def load_bar(**lister_args):
    """Load bar"""
    pass


@shared_task(name=__name__ + ".ping")
def _ping():
    return "OK"



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, Tuple

import attr
import pytest

from swh.scheduler.cli.utils import lister_task_type
from swh.scheduler.interface import SchedulerInterface
from swh.scheduler.tests.common import TASK_TYPES
from swh.scheduler.tests.test_cli import invoke as basic_invoke


def invoke(scheduler, args: Tuple[str, ...] = (), catch_exceptions: bool = False):
    return basic_invoke(
        scheduler, args=["add-forge-now", *args], catch_exceptions=catch_exceptions
    )


def test_schedule_first_visits_cli_unknown_visit_type(
    swh_scheduler,
):
    "Calling cli without a known visit type should raise"
    result = invoke(
        swh_scheduler,
        (
            "schedule-first-visits",
            "-t",
            "unknown-vt0",
            "--type-name",
            "unknown-visit-type1",
        ),
        True,
    )
    assert "Unknown" in result.output
    assert result.exit_code != 0


@pytest.mark.parametrize(
    "cmd_args, subcmd_args",
    [
        ([], []),
        ([], ["--lister-name", "github", "--lister-instance-name", "github"]),
        (["--preset", "staging"], []),
        ([], ["--queue-name-prefix", "add_forge_now_slow"]),
    ],
)
def test_schedule_first_visits_cli(
    mocker,
    swh_scheduler,
    swh_scheduler_celery_app,
    listed_origins_by_type,
    cmd_args,
    subcmd_args,
):
    for task_type in TASK_TYPES.values():
        swh_scheduler.create_task_type(task_type)

    visit_type = next(iter(listed_origins_by_type))

    # enabled origins by default except when --staging flag is provided
    enabled = "staging" not in cmd_args

    for origins in listed_origins_by_type.values():
        swh_scheduler.record_listed_origins(
            (attr.evolve(o, enabled=enabled) for o in origins)
        )

    get_queue_length = mocker.patch(
        "swh.scheduler.celery_backend.config.get_queue_length"
    )
    get_queue_length.return_value = None

    send_task = mocker.patch.object(swh_scheduler_celery_app, "send_task")
    send_task.return_value = None

    command_args = (
        cmd_args + ["schedule-first-visits", "--type-name", visit_type] + subcmd_args
    )

    result = invoke(swh_scheduler, args=tuple(command_args))
    assert result.exit_code == 0

    scheduled_tasks = {
        (call[0][0], call[1]["kwargs"]["url"]) for call in send_task.call_args_list
    }

    expected_tasks = {
        (TASK_TYPES[origin.visit_type]["backend_name"], origin.url)
        for origin in listed_origins_by_type[visit_type]
    }

    assert scheduled_tasks == expected_tasks

    # Ensure the scheduling happens on the right queue
    queue_name = send_task.call_args[1]["queue"]
    queue_name_prefix, queue_name_suffix = queue_name.split(":")

    try:
        arg_index = subcmd_args.index("--queue-name-prefix")
        expected_queue_name_prefix = subcmd_args[arg_index + 1]
    except ValueError:
        expected_queue_name_prefix = "add_forge_now"

    assert queue_name_prefix == expected_queue_name_prefix
    assert queue_name_suffix == "swh.loader.git.tasks.UpdateGitRepository"


def _create_task_type(
    swh_scheduler: SchedulerInterface, lister_name: str, listing_type: str = "full"
) -> Dict:
    task_type = {
        "type": lister_task_type(lister_name, listing_type),  # only relevant bit
        "description": f"{listing_type} listing",
        "backend_name": "swh.example.backend",
        "default_interval": "1 day",
        "min_interval": "1 day",
        "max_interval": "1 day",
        "backoff_factor": "1",
        "max_queue_length": "100",
        "num_retries": 3,
    }
    swh_scheduler.create_task_type(task_type)
    task_type = swh_scheduler.get_task_type(task_type["type"])
    assert task_type is not None
    return task_type


@pytest.mark.parametrize("preset", ["staging", "production"])
def test_schedule_register_lister(swh_scheduler, stored_lister, preset):
    # given
    assert stored_lister is not None
    lister_name = stored_lister.name
    # Let's create all possible associated lister task types
    full = _create_task_type(swh_scheduler, lister_name, "full")
    incremental = _create_task_type(swh_scheduler, lister_name, "incremental")

    # Let's trigger the registering of that lister
    result = invoke(
        swh_scheduler,
        [
            "--preset",
            preset,
            "register-lister",
            lister_name,
            "url=https://example.org",
        ],
    )

    output = result.output.lstrip()

    expected_msgs = []
    if preset == "production":
        # 2 tasks: 1 full + 1 incremental (tomorrow) with recurring policy
        expected_msgs = ["Policy: recurring", incremental["type"], "Next run: tomorrow"]
    else:
        # 1 task full with policy oneshot
        expected_msgs = ["Policy: oneshot"]

    # In any case, there is the full listing type too
    expected_msgs.append(full["type"])

    assert len(expected_msgs) > 0
    for msg in expected_msgs:
        assert msg in output


def test_register_lister_unknown_task_type(swh_scheduler):
    """When scheduling unknown task type, the cli should raise."""
    result = invoke(
        swh_scheduler,
        [
            "register-lister",
            "unknown-lister-type-should-raise",
        ],
        True,
    )
    assert "Unknown" in result.output
    assert result.exit_code != 0
